%% Adaptive inverse scale space method
%
% This is an implementation of the adaptive inverse scale space method
% (Algorithm 2 proposed in [1]). 
%
% Author:   Martin Benning - mailto:mb941@cam.ac.uk
%           Michael Moeller
%
% Date: 14.12.2016
%
% Input arguments:              A - data matrix
%                               f - data vector
% 
% Optional input arguments:     ploton - plot enabled/disabled
%                               maxiter - maximum no. of iterations
%                               threshold - threshold for discrep. princ.
%
% Output:                       u - vector that solves || Au - f || <= 
%                                   threshold                       
%                               
% Examples:     u = adaptinvscalespace(A, f);
%               u = adaptinvscalespace(A, f, true, 100, 10^-3);
%
function u = adaptinvscalespace(A, f, ploton, maxiter, threshold)

if nargin > 5
    error('Too many input arguments.')
elseif nargin < 2
    error('Not enough input arguments.')
end
if nargin < 5
    threshold = 10^-6;
end
if nargin < 4
    maxiter = 1000;
end
if nargin < 3
    ploton = false;
end

if ploton == true
    fig = figure(); % create empty figure if output is enabled
end

m = size(A,2); % dimension of output argument u
Atansf = A' * f; % A^T f
numerr = 10^(-8); % numerical error tolerance for indexset indentification
t1 = 1 ./ max(abs(Atansf)); % first time step
p = t1 * Atansf; % first subgradient
I = (abs(p) > 1 - numerr); % index set 
k = 1;
stopcrit = false;
while (k <= maxiter) && ~stopcrit
    
    % solve the least squares problem only with the components in I and
    % correct sign
    B = full(A(:, logical(I))); 
    temp = sign(p(I)');
    B = B .* repmat(temp, [size(B, 1), 1]); %change signs of the columns of B  
    utemp = lsqnonneg(B, f); % solve non-negative least squares problem
    utemp = utemp .* temp';  % undo the change of sign
    
    % store result in u %
    u = zeros(m, 1);
    u(I) = utemp;
    
    % if output is enabled, plot intermediate result and pause
    if ploton == true
        figure(fig)        
        axes1 = axes('Parent', figure(fig));        
        hold(axes1,'on');
        [x,y] = peakvectors(u, 1);
        line1 = line(x, y, 'Color', 'b', 'LineWidth', 2);
        plot1 = plot(1:length(u), p, 'Color', 'r', 'LineWidth', 2);
        set(line1, 'DisplayName', ['aISS solution at iteration ', ...
            num2str(k)]);
        set(plot1, 'DisplayName', ['Subgradient at iteration ', ...
            num2str(k)], 'LineStyle', '--');
        axis([1 length(u) (min([u' p']) - 0.4) (max([u' p']) + 0.4)])
        a = min(ceil(min([u' p']) - 0.4), -2);
        b = max(floor(max([u' p']) + 0.4), 2);
        h = gca;
        h.YTick = [a, -1, 0, 1, b];
        legend1 = legend(axes1, 'show');
        box(axes1, 'on');
        set(axes1, 'FontSize', 22);
        set(legend1, 'Location', 'best');
        pause;
    end

    
    % determine the next time step t
    
    residual = (Atansf - A'*(A*u)); %determine residual    
    zeroresidual = (abs(residual) < 10^(-10)); % omit residuals that are 
    % essentially zero
    residualtemp = residual(logical(1 - zeroresidual));
    ptemp = p(logical(1 - zeroresidual));         
    c = ptemp - t1 * residualtemp;     
    ttemp1 = (1 - c) ./ residualtemp;
    
    ttemp1((ttemp1 < t1 + numerr)) = []; % discard times which are smaller 
    % than t^k    
    ttemp2 = (-1 - c) ./ residualtemp;    
    ttemp2((ttemp2 < t1 + numerr)) = []; % discard times which are smaller 
    % than t^k
    
    % stop if there are no more time steps to compute; otherwise proceed
    if isempty(ttemp1) && isempty(ttemp2)
        stopcrit = true;
    else        
        t2 = min([min(ttemp1), min(ttemp2)]); % find the minimum time among
        % all possible times      
        p = p + (t2 - t1) * residual; % update the subgradient                
        I = (abs(p) > (1 - numerr)); % update index set
        t1 = t2;
    end
    if norm(A*u - f, 2) < threshold % stop if discrep. princ. is violated
        stopcrit = true;
    end
    k = k + 1;
end

if ploton == true
    close(fig)
end

k = k - 1;
if k == maxiter
    disp(['The algorithm was stopped after ' num2str(maxiter) ...
        ' iterations and might not have fully converged.']);
end
%% References
%
% [1] An adaptive inverse scale space method for compressed sensing. Martin
% Burger, Michael Moeller, Martin Benning, Stanley Osher. Mathematics of
% Computation, 2013.